import abc
import collections
import enum
import functools
import typing
import struct
import numpy as np

from ndsi import StreamError
from ndsi.frame import JPEGFrame, H264Frame, FrameFactory
from ndsi.frame import VIDEO_FRAME_FORMAT_H264, VIDEO_FRAME_FORMAT_MJPEG


__all__ = [
    "DataFormat",
    "DataFormatter",
    "DataMessage",
    "VideoDataFormatter",
    "VideoValue",
    "GazeDataFormatter",
    "GazeValue",
    "AnnotateDataFormatter",
    "AnnotateValue",
    "IMUDataFormatter",
    "IMUValue",
]


NANO = 1e-9


"""
To add a new data format version, in `formatter.py`:
1. Add a new case to the `DataFormat` enum.
2. For each concrete sublcass of `DataFormatter` (except `UnsupportedFormatter`), extend the implementation of `get_formatter` to correctly handle the new format version.
3. Run the test suit to make sure that all the tests pass again.
4. Write additional tests to cover the custom behaviour of the new data format.
"""


@enum.unique
class DataFormat(enum.Enum):
    """
    `DataFormat` enum represents the format for serializing and deserializing data between NDSI hosts and clients.
    """

    V3 = "v3"
    V4 = "v4"

    @staticmethod
    def latest() -> "DataFormat":
        return max(DataFormat.supported_formats(), key=lambda f: f.version_major)

    @staticmethod
    def supported_formats() -> typing.Set["DataFormat"]:
        return set(DataFormat)

    @property
    def version_major(self) -> int:
        return int(self.value[1:])

    def __str__(self) -> str:
        return self.value


class DataMessage(typing.NamedTuple):
    sensor_id: str
    header: bytes
    body: bytes


DataValue = typing.TypeVar("DataValue")


class DataFormatter(typing.Generic[DataValue], abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def get_formatter(format: DataFormat) -> "DataFormatter":
        pass

    @abc.abstractmethod
    def encode_msg(self, value: DataValue) -> DataMessage:
        pass

    @abc.abstractmethod
    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[DataValue]:
        pass


##########


class UnsupportedFormatter(DataFormatter[typing.Any]):
    """
    Represents a formatter that is not supported for a specific data format and sensor type combination.
    """

    @staticmethod
    def get_formatter(format: DataFormat) -> "UnsupportedFormatter":
        return UnsupportedFormatter()

    def encode_msg(self, value: DataValue) -> DataMessage:
        raise ValueError("Unsupported data format.")

    def decode_msg(self, value: DataMessage) -> typing.Iterator[DataValue]:
        raise ValueError("Unsupported data format.")


##########


VideoValue = typing.Union[JPEGFrame, H264Frame]


class VideoDataFormatter(DataFormatter[VideoValue]):
    def __init__(self):
        super().__init__()
        self._frame_factory = FrameFactory()
        self._newest_h264_frame = None

    def reset(self):
        self._newest_h264_frame = None

    @staticmethod
    @functools.lru_cache(maxsize=1, typed=True)
    def get_formatter(
        format: DataFormat,
    ) -> typing.Union["VideoDataFormatter", UnsupportedFormatter]:
        if format == DataFormat.V3:
            return _VideoDataFormatter_V3()
        if format == DataFormat.V4:
            return _VideoDataFormatter_V4()
        raise ValueError(format)

    def encode_msg(self, value: VideoValue) -> DataMessage:
        raise NotImplementedError()


class _VideoDataFormatter_V3(VideoDataFormatter):
    def decode_msg(self, data_msg: DataMessage) -> VideoValue:
        meta_data = struct.unpack("<LLLLdLL", data_msg.header)
        meta_data_mutable = list(meta_data)
        meta_data_mutable[4] *= 1e6  #  Convert timestamp s -> us
        meta_data = tuple(meta_data_mutable)
        if meta_data[0] == VIDEO_FRAME_FORMAT_MJPEG:
            yield self._frame_factory.create_jpeg_frame(data_msg.body, meta_data)
        elif meta_data[0] == VIDEO_FRAME_FORMAT_H264:
            frame = self._frame_factory.create_h264_frame(data_msg.body, meta_data)
            self._newest_h264_frame = frame or self._newest_h264_frame
            yield self._newest_h264_frame
        else:
            raise StreamError("Frame was not of format MJPEG or H264")


class _VideoDataFormatter_V4(VideoDataFormatter):
    def decode_msg(self, data_msg: DataMessage) -> VideoValue:
        meta_data = struct.unpack("<LLLLQLL", data_msg.header)
        meta_data_mutable = list(meta_data)
        meta_data_mutable[4] /= 1e3  #  Convert timestamp ns -> us
        meta_data = tuple(meta_data_mutable)
        if meta_data[0] == VIDEO_FRAME_FORMAT_MJPEG:
            yield self._frame_factory.create_jpeg_frame(data_msg.body, meta_data)
        elif meta_data[0] == VIDEO_FRAME_FORMAT_H264:
            frame = self._frame_factory.create_h264_frame(data_msg.body, meta_data)
            self._newest_h264_frame = frame or self._newest_h264_frame
            yield self._newest_h264_frame
        else:
            raise StreamError("Frame was not of format MJPEG or H264")


##########


# TODO: Where is the format for annotation data defined?
class AnnotateValue(typing.NamedTuple):
    key: int
    timestamp: float


class AnnotateDataFormatter(DataFormatter[AnnotateValue]):
    @staticmethod
    @functools.lru_cache(maxsize=1, typed=True)
    def get_formatter(
        format: DataFormat,
    ) -> typing.Union["AnnotateDataFormatter", UnsupportedFormatter]:
        if format == DataFormat.V3:
            return _AnnotateDataFormatter_V3()
        if format == DataFormat.V4:
            return _AnnotateDataFormatter_V4()
        raise ValueError(format)

    def encode_msg(self, value: AnnotateValue) -> DataMessage:
        raise NotImplementedError()


class _AnnotateDataFormatter_V3(AnnotateDataFormatter):
    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[AnnotateValue]:
        # NOTE: Annotation sensor is currently not NDSI-conformant.
        key, ts = struct.unpack("<Bd", data_msg[0])
        yield AnnotateValue(key=key, timestamp=ts)


class _AnnotateDataFormatter_V4(AnnotateDataFormatter):
    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[AnnotateValue]:
        # NOTE: Annotation sensor is currently not NDSI-conformant.
        key, ts = struct.unpack("<BQ", data_msg[0])
        ts *= NANO
        yield AnnotateValue(key=key, timestamp=ts)


##########


class GazeValue(typing.NamedTuple):
    x: int
    y: int
    timestamp: float


class GazeDataFormatter(DataFormatter[GazeValue]):
    @staticmethod
    @functools.lru_cache(maxsize=1, typed=True)
    def get_formatter(
        format: DataFormat,
    ) -> typing.Union["GazeDataFormatter", UnsupportedFormatter]:
        if format == DataFormat.V3:
            return UnsupportedFormatter()
        if format == DataFormat.V4:
            return _GazeDataFormatter_V4()
        raise ValueError(format)

    def encode_msg(self, value: GazeValue) -> DataMessage:
        raise NotImplementedError()


class _GazeDataFormatter_V4(GazeDataFormatter):
    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[GazeValue]:
        (ts,) = struct.unpack("<Q", data_msg.header)
        ts *= NANO
        x, y = struct.unpack("<ff", data_msg.body)
        yield GazeValue(x=x, y=y, timestamp=ts)


##########


class IMUValue(typing.NamedTuple):
    timestamp: float
    accel_x: float
    accel_y: float
    accel_z: float
    gyro_x: float
    gyro_y: float
    gyro_z: float


class IMUDataFormatter(DataFormatter[IMUValue]):
    @staticmethod
    @functools.lru_cache(maxsize=1, typed=True)
    def get_formatter(
        format: DataFormat,
    ) -> typing.Union["IMUDataFormatter", UnsupportedFormatter]:
        if format == DataFormat.V3:
            return _IMUDataFormatter_V3()
        if format == DataFormat.V4:
            return _IMUDataFormatter_V4()
        raise ValueError(format)

    def encode_msg(self, value: IMUValue) -> DataMessage:
        raise NotImplementedError()


class _IMUDataFormatter_V3(IMUDataFormatter):
    CONTENT_DTYPE = np.dtype(
        [
            ("time_s", "<f8"),
            ("accel_x", "<f4"),
            ("accel_y", "<f4"),
            ("accel_z", "<f4"),
            ("gyro_x", "<f4"),
            ("gyro_y", "<f4"),
            ("gyro_z", "<f4"),
        ]
    )

    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[IMUValue]:
        content = np.frombuffer(data_msg.body, dtype=self.CONTENT_DTYPE).view(
            np.recarray
        )
        for imu_frame in content:
            yield IMUValue(*imu_frame)


class _IMUDataFormatter_V4(IMUDataFormatter):
    CONTENT_DTYPE = np.dtype(
        [
            ("time_ns", "<u8"),
            ("accel_x", "<f4"),
            ("accel_y", "<f4"),
            ("accel_z", "<f4"),
            ("gyro_x", "<f4"),
            ("gyro_y", "<f4"),
            ("gyro_z", "<f4"),
        ]
    )

    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[IMUValue]:
        content = np.frombuffer(data_msg.body, dtype=self.CONTENT_DTYPE).view(
            np.recarray
        )
        for imu_frame in content:
            ts, *data = imu_frame
            ts *= NANO
            yield IMUValue(ts, *data)


##########


class EventValue(typing.NamedTuple):
    timestamp: float
    label: str


class EventDataFormatter(DataFormatter[EventValue]):
    @staticmethod
    @functools.lru_cache(maxsize=1, typed=True)
    def get_formatter(
        format: DataFormat,
    ) -> typing.Union["EventDataFormatter", UnsupportedFormatter]:
        if format == DataFormat.V3:
            return UnsupportedFormatter()
        if format == DataFormat.V4:
            return _EventDataFormatter_V4()
        raise ValueError(format)

    def encode_msg(self, value: EventValue) -> DataMessage:
        raise NotImplementedError()


class _EventDataFormatter_V4(EventDataFormatter):

    _encoding_lookup = {0: "utf-8"}

    def decode_msg(self, data_msg: DataMessage) -> typing.Iterator[EventValue]:
        """
        1. sensor UUID
        2. header:
            - int_64 timestamp_le
            - uint32 body_length_le
            - uint32 encoding_le
                = 0 -> "utf-8"
        3. body:
            - `encoding_le` encoded string of length `body_length_le`
        """
        ts, len_, enc_code = struct.unpack("<qii", data_msg.header)
        ts *= NANO
        enc = self._encoding_lookup[enc_code]
        body = data_msg.body.bytes[:len_]
        label = body.decode(enc)
        yield EventValue(label=label, timestamp=ts)



"""
(*)~----------------------------------------------------------------------------------
 Pupil - eye tracking platform
 Copyright (C) 2012-2015  Pupil Labs

 Distributed under the terms of the CC BY-NC-SA License.
 License details are in the file LICENSE, distributed as part of this software.
----------------------------------------------------------------------------------~(*)
"""

import abc
import collections
import functools
import itertools
import json as serial
import logging
import sys
import time
import traceback as tb
import typing

import zmq
from pyre import Pyre, PyreEvent, zhelper

from ndsi import __protocol_version__

from ndsi.formatter import DataFormat
from ndsi.sensor import SensorType, Sensor


logger = logging.getLogger(__name__)


NetworkEvent = typing.Mapping[str, typing.Any]


NetworkEventCallback = typing.Callable[[typing.Any, NetworkEvent], None]


NetworkSensor = typing.Mapping[str, typing.Any]


class NetworkInterface(abc.ABC):
    """
    Public interface for a Network-like object.
    """

    @property
    @abc.abstractmethod
    def has_events(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def running(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def sensors(self) -> typing.Mapping[str, NetworkSensor]:
        pass

    @property
    @abc.abstractmethod
    def callbacks(self) -> typing.Iterable[NetworkEventCallback]:
        pass

    @callbacks.setter
    @abc.abstractmethod
    def callbacks(self, value: typing.Iterable[NetworkEventCallback]):
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def whisper(self, peer, msg_p):
        """Send message to single peer, specified as a UUID string"""
        pass

    @abc.abstractmethod
    def rejoin(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    @abc.abstractmethod
    def handle_event(self):
        pass

    @abc.abstractmethod
    def sensor(
        self, sensor_uuid: str, callbacks: typing.Iterable[NetworkEventCallback] = ()
    ) -> Sensor:
        pass


class _NetworkNode(NetworkInterface):
    """
    Communication node

    Creates Pyre node and handles all communication.
    """

    def __init__(
        self, format: DataFormat, context=None, name=None, headers=(), callbacks=()
    ):
        self._name = name
        self._format = format
        self._headers = headers
        self._pyre_node = None
        self._context = context or zmq.Context()
        self._sensors_by_host = {}
        self._callbacks = [self._on_event] + list(callbacks)

    # Public NetworkInterface API

    @property
    def has_events(self) -> bool:
        return self.running and self._pyre_node.socket().get(zmq.EVENTS) & zmq.POLLIN

    @property
    def running(self) -> bool:
        return bool(self._pyre_node)

    @property
    def sensors(self) -> typing.Mapping[str, NetworkSensor]:
        sensors = {}
        for sensor in self._sensors_by_host.values():
            sensors.update(sensor)
        return sensors

    @property
    def callbacks(self) -> typing.Iterable[NetworkEventCallback]:
        return self._callbacks

    @callbacks.setter
    def callbacks(self, value: typing.Iterable[NetworkEventCallback]):
        self._callbacks = value

    def start(self):
        # Setup node
        logger.debug("Starting network...")
        self._pyre_node = Pyre(self._name)
        self._name = self._pyre_node.name()
        for header in self._headers:
            self._pyre_node.set_header(*header)
        self._pyre_node.join(self._group)
        self._pyre_node.start()

    def whisper(self, peer, msg_p):
        if self._format == DataFormat.V3:
            return  # no-op
        elif self._format == DataFormat.V4:
            self._pyre_node.whisper(peer, msg_p)
        else:
            raise NotImplementedError()

    def rejoin(self):
        for sensor_uuid, sensor in list(self.sensors.items()):
            self._execute_callbacks(
                {
                    "subject": "detach",
                    "sensor_uuid": sensor_uuid,
                    "sensor_name": sensor["sensor_name"],
                    "host_uuid": sensor["host_uuid"],
                    "host_name": sensor["host_name"],
                }
            )
        self._pyre_node.leave(self._group)
        self._pyre_node.join(self._group)

    def stop(self):
        logger.debug("Stopping network...")
        self._pyre_node.leave(self._group)
        self._pyre_node.stop()
        self._pyre_node = None

    def handle_event(self):
        if not self.has_events:
            return
        event = PyreEvent(self._pyre_node)
        uuid = event.peer_uuid
        if event.type == "SHOUT" or event.type == "WHISPER":
            try:
                payload = event.msg.pop(0).decode()
                msg = serial.loads(payload)
                msg["subject"]
                msg["sensor_uuid"]
                msg["host_uuid"] = event.peer_uuid.hex
                msg["host_name"] = event.peer_name
            except serial.decoder.JSONDecodeError:
                logger.warning('Malformatted message: "{}"'.format(payload))
            except (ValueError, KeyError):
                logger.warning("Malformatted message: {}".format(msg))
            except Exception:
                logger.debug(tb.format_exc())
            else:
                if msg["subject"] == "attach":
                    if self.sensors.get(msg["sensor_uuid"]):
                        # Sensor already attached. Drop event
                        return
                    sensor_type = SensorType.supported_sensor_type_from_str(
                        msg["sensor_type"]
                    )
                    if sensor_type is None:
                        logger.debug(
                            "Unsupported sensor type: {}".format(msg["sensor_type"])
                        )
                        return
                elif msg["subject"] == "detach":
                    sensor_entry = self.sensors.get(msg["sensor_uuid"])
                    # Check if sensor has been detached already
                    if not sensor_entry:
                        return
                    msg.update(sensor_entry)
                else:
                    logger.debug("Unknown host message: {}".format(msg))
                    return
                self._execute_callbacks(msg)
        elif event.type == "JOIN":
            # possible values for `group_version`
            # - [<unrelated group>]
            # - [<unrelated group>, <unrelated version>]
            # - ['pupil-mobile']
            # - ['pupil-mobile', <version>]
            group_version = event.group.split("-v")
            group = group_version[0]
            version = group_version[1] if len(group_version) > 1 else "0"

        elif event.type == "EXIT":
            gone_peer = event.peer_uuid.hex
            for host_uuid, sensors in list(self._sensors_by_host.items()):
                if host_uuid != gone_peer:
                    continue
                for sensor_uuid, sensor in list(sensors.items()):
                    self._execute_callbacks(
                        {
                            "subject": "detach",
                            "sensor_uuid": sensor_uuid,
                            "sensor_name": sensor["sensor_name"],
                            "host_uuid": host_uuid,
                            "host_name": sensor["host_name"],
                        }
                    )
        else:
            logger.debug("Dropping {}".format(event))

    def sensor(
        self, sensor_uuid: str, callbacks: typing.Iterable[NetworkEventCallback] = ()
    ) -> Sensor:
        try:
            sensor_settings = self.sensors[sensor_uuid].copy()
        except KeyError:
            raise ValueError('"{}" is not an available sensor id.'.format(sensor_uuid))

        sensor_type_str = sensor_settings.pop("sensor_type", "unknown")
        sensor_type = SensorType.supported_sensor_type_from_str(sensor_type_str)

        if sensor_type is None:
            raise ValueError(
                'Sensor of type "{}" is not supported.'.format(sensor_type_str)
            )

        return Sensor.create_sensor(
            sensor_type=sensor_type,
            format=self._format,
            context=self._context,
            callbacks=callbacks,
            **sensor_settings,
        )

    # Public

    def __str__(self):
        return "<{} {} [{}]>".format(__name__, self._name, self._pyre_node.uuid().hex)

    # Private

    @property
    def _group(self) -> str:
        return group_name_from_format(self._format)

    def _execute_callbacks(self, event):
        for callback in self.callbacks:
            callback(self, event)

    def _on_event(self, caller, event):
        if event["subject"] == "attach":
            subject_less = event.copy()
            del subject_less["subject"]
            host_uuid = event["host_uuid"]
            host_sensor = {event["sensor_uuid"]: subject_less}
            try:
                self._sensors_by_host[host_uuid].update(host_sensor)
            except KeyError:
                self._sensors_by_host[host_uuid] = host_sensor
            logger.debug(f'Attached {host_uuid}.{event["sensor_uuid"]}')
        elif event["subject"] == "detach":
            for host_uuid, sensors in self._sensors_by_host.items():
                try:
                    del sensors[event["sensor_uuid"]]
                    logger.debug(f'Detached {host_uuid}.{event["sensor_uuid"]}')
                except KeyError:
                    pass
            hosts_to_remove = [
                host_uuid
                for host_uuid, sensors in self._sensors_by_host.items()
                if len(sensors) == 0
            ]
            for host_uuid in hosts_to_remove:
                del self._sensors_by_host[host_uuid]


class Network(NetworkInterface):
    def __init__(
        self,
        formats: typing.Set[DataFormat] = None,
        context=None,
        name=None,
        headers=(),
        callbacks=(),
    ):
        formats = formats or {DataFormat.latest()}
        self.context = context or zmq.Context()
        self._callbacks = callbacks
        self._nodes = [
            _NetworkNode(
                format=format,
                context=self.context,
                name=name,
                headers=headers,
                callbacks=self._callbacks,
            )
            for format in formats
        ]
        assert len(self._nodes) > 0

    # Public NetworkInterface API

    @property
    def has_events(self) -> bool:
        return any(node.has_events for node in self._nodes)

    @property
    def running(self) -> bool:
        return any(node.running for node in self._nodes)

    @property
    def sensors(self) -> typing.Mapping[str, NetworkSensor]:
        sensors = collections.ChainMap(*(n.sensors for n in self._nodes))
        return dict(sensors)

    @property
    def callbacks(self) -> typing.Iterable[NetworkEventCallback]:
        return self._callbacks

    @callbacks.setter
    def callbacks(self, value: typing.Iterable[NetworkEventCallback]):
        self._callbacks = value
        for node in self._nodes:
            node.callbacks = value

    def start(self):
        for node in self._nodes:
            node.start()

    def whisper(self, peer, msg_p):
        for node in self._nodes:
            node.whisper(peer=peer, msg_p=msg_p)

    def rejoin(self):
        for node in self._nodes:
            node.rejoin()

    def stop(self):
        for node in self._nodes:
            node.stop()

    def handle_event(self):
        for node in self._nodes:
            node.handle_event()

    def sensor(
        self, sensor_uuid: str, callbacks: typing.Iterable[NetworkEventCallback] = ()
    ) -> Sensor:
        for node in self._nodes:
            if sensor_uuid in node.sensors:
                return node.sensor(sensor_uuid=sensor_uuid, callbacks=callbacks)
        raise ValueError('"{}" is not an available sensor id.'.format(sensor_uuid))


def group_name_from_format(format: DataFormat) -> str:
    return "pupil-mobile-{}".format(format)



"""
(*)~----------------------------------------------------------------------------------
 Pupil - eye tracking platform
 Copyright (C) 2012-2015  Pupil Labs

 Distributed under the terms of the CC BY-NC-SA License.
 License details are in the file LICENSE, distributed as part of this software.
----------------------------------------------------------------------------------~(*)
"""


class CaptureError(Exception):
    def __init__(self, message):
        super(CaptureError, self).__init__()
        self.message = message


class StreamError(CaptureError):
    def __init__(self, message):
        super(StreamError, self).__init__(message)
        self.message = message


from ndsi.formatter import DataFormat


__version__ = "1.4.2"
__protocol_version__ = str(DataFormat.latest().version_major)


from ndsi.network import Network
from ndsi.sensor import Sensor
from ndsi.writer import H264Writer

from ndsi import frame



"""
(*)~----------------------------------------------------------------------------------
 Pupil - eye tracking platform
 Copyright (C) 2012-2015  Pupil Labs

 Distributed under the terms of the CC BY-NC-SA License.
 License details are in the file LICENSE, distributed as part of this software.
----------------------------------------------------------------------------------~(*)
"""

import abc
import enum
import json as serial
import traceback as tb
import numpy as np
import zmq

import logging

logger = logging.getLogger(__name__)

from ndsi import StreamError

import typing

from ndsi.formatter import DataFormatter, DataFormat, DataMessage
from ndsi.formatter import VideoDataFormatter, VideoValue
from ndsi.formatter import GazeDataFormatter, GazeValue
from ndsi.formatter import AnnotateDataFormatter, AnnotateValue
from ndsi.formatter import IMUDataFormatter, IMUValue
from ndsi.formatter import EventDataFormatter, EventValue


NANO = 1e-9


class NotDataSubSupportedError(Exception):
    def __init__(self, value=None):
        self.value = value or "This sensor does not support data subscription."

    def __str__(self):
        return repr(self.value)


"""
To add a new sensor, in `sensor.py`:
1. Add a new case to the `SensorType` enum.
2. Add a new subclass of `Sensor` and implement the custom sensor behaviour.
3. Add a new entry into the `_SENSOR_TYPE_CLASS_MAP`, mapping the new `SensorType` case to the new `Sensor` subclass.
4. (Optional) If the new `Sensor` subclass will include `SensorFetchDataMixin`, the subclass must define a property `formatter` that returns an instance of `DataFormatter` which serializes/deserializes the data handled by the sensor.
5. Run the test suit to make sure that all the tests pass again.
6. Write additional tests to cover the custom behaviour of the new sensor type.
"""


@enum.unique
class SensorType(enum.Enum):
    HARDWARE = "hardware"
    VIDEO = "video"
    ANNOTATE = "annotate"
    GAZE = "gaze"
    IMU = "imu"
    EVENT = "event"
    LED = "led"

    @staticmethod
    def supported_types() -> typing.Set["SensorType"]:
        return set(SensorType)

    @staticmethod
    def supported_sensor_type_from_str(
        sensor_type_name: str,
    ) -> typing.Optional["SensorType"]:
        try:
            sensor_type = SensorType(sensor_type_name)
        except ValueError:
            return None
        if sensor_type not in SensorType.supported_types():
            return None
        return sensor_type

    def __str__(self) -> str:
        return self.value


class Sensor:
    @staticmethod
    def class_for_type(sensor_type: SensorType):
        try:
            return _SENSOR_TYPE_CLASS_MAP[sensor_type]
        except KeyError:
            raise ValueError("Unknown sensor type: {}".format(sensor_type))

    @staticmethod
    def create_sensor(sensor_type: SensorType, **kwargs) -> "Sensor":
        sensor_class = Sensor.class_for_type(sensor_type=sensor_type)
        # TODO: Passing sensor_type to the class init as str, to preserve API compatibility.
        #       Ideally, the sensor_type passed and stored by Sensor is of type SensorType.
        kwargs["sensor_type"] = str(sensor_type)
        return sensor_class(**kwargs)

    def __init__(
        self,
        format: DataFormat,
        host_uuid,
        host_name,
        sensor_uuid,
        sensor_name,
        sensor_type,
        notify_endpoint,
        command_endpoint,
        data_endpoint=None,
        context=None,
        callbacks=(),
    ):
        self.format = format
        self.callbacks = [self.on_notification] + list(callbacks)
        self.context = context or zmq.Context()
        self.host_uuid = host_uuid
        self.host_name = host_name
        self.name = sensor_name
        self.type = sensor_type
        self.uuid = sensor_uuid
        self.notify_endpoint = notify_endpoint
        self.command_endpoint = command_endpoint
        self.data_endpoint = data_endpoint
        self.controls: typing.Dict[str, typing.Any] = {}

        self.notify_sub = context.socket(zmq.SUB)
        self.notify_sub.connect(self.notify_endpoint)
        self.notify_sub.subscribe(self.uuid)

        self.command_push = context.socket(zmq.PUSH)
        self.command_push.connect(self.command_endpoint)

        self._init_data_sub(context)

        self.refresh_controls()

    def _init_data_sub(self, context):
        if self.data_endpoint:
            self.data_sub = context.socket(zmq.SUB)
            self.data_sub.set_hwm(3)
            self.data_sub.connect(self.data_endpoint)
            self.data_sub.subscribe(self.uuid)
        else:
            self.data_sub = None

    def unlink(self):
        self.notify_sub.unsubscribe(self.uuid)
        self.notify_sub.close(linger=0)
        self.command_push.close(linger=0)
        if self.supports_data_subscription:
            self.data_sub.unsubscribe(self.uuid)
            self.data_sub.close(linger=0)

    @property
    def supports_data_subscription(self):
        return bool(self.data_sub)

    @property
    def has_notifications(self):
        has_n = self.notify_sub.get(zmq.EVENTS) & zmq.POLLIN
        return has_n

    @property
    def has_data(self):
        try:
            return self.data_sub.get(zmq.EVENTS) & zmq.POLLIN
        except AttributeError:
            raise NotDataSubSupportedError()

    def __str__(self):
        return "<{} {}@{} [{}]>".format(__name__, self.name, self.host_name, self.type)

    def handle_notification(self):
        raw_notification = self.notify_sub.recv_multipart()
        if len(raw_notification) != 2:
            logger.debug(
                "Message for sensor {} has not correct amount of frames: {}".format(
                    self.uuid, raw_notification
                )
            )
            return
        sender_id = raw_notification[0].decode()
        notification_payload = raw_notification[1].decode()
        try:
            if sender_id != self.uuid:
                raise ValueError(
                    "Message was destined for {} but was recieved by {}".format(
                        sender_id, self.uuid
                    )
                )
            notification = serial.loads(notification_payload)
            notification["subject"]
        except serial.decoder.JSONDecodeError:
            logger.debug(
                "JSONDecodeError for payload: `{}`".format(notification_payload)
            )
        except Exception:
            logger.debug(tb.format_exc())
        else:
            try:
                self.execute_callbacks(notification)
            except:
                logger.debug(tb.format_exc())

    def execute_callbacks(self, event):
        for callback in self.callbacks:
            callback(self, event)

    def on_notification(self, caller, notification):
        if notification["subject"] == "update":

            class UnsettableDict(dict):
                def __getitem__(self, key):
                    return self.get(key)

                def __setitem__(self, key, value):
                    raise ValueError(
                        "Dictionary is read-only. Use Sensor.set_control_value instead."
                    )

            ctrl_id_key = notification["control_id"]
            if ctrl_id_key in self.controls:
                self.controls[ctrl_id_key].update(
                    UnsettableDict(notification["changes"])
                )
            else:
                self.controls[ctrl_id_key] = UnsettableDict(notification["changes"])
        elif notification["subject"] == "remove":
            try:
                del self.controls[notification["control_id"]]
            except KeyError:
                pass

    def get_data(self, copy=True):
        try:
            return self.data_sub.recv_multipart(copy=copy)
        except AttributeError:
            raise NotDataSubSupportedError()

    def refresh_controls(self):
        cmd = serial.dumps({"action": "refresh_controls"})
        self.command_push.send_string(self.uuid, flags=zmq.SNDMORE)
        self.command_push.send_string(cmd)

    def reset_all_control_values(self):
        for control_id in self.controls:
            self.reset_control_value(control_id)

    def reset_control_value(self, control_id):
        if control_id in self.controls:
            if "def" in self.controls[control_id]:
                value = self.controls[control_id]["def"]
                self.set_control_value(control_id, value)
            else:
                logger.error(
                    (
                        "Could not reset control `{}` because it does not have a default value."
                    ).format(control_id)
                )
        else:
            logger.error("Could not reset unknown control `{}`".format(control_id))

    def set_control_value(self, control_id, value):
        try:
            dtype = self.controls[control_id]["dtype"]
            if dtype == "bool":
                value = bool(value)
            elif dtype == "string":
                value = str(value)
            elif dtype == "integer":
                value = int(value)
            elif dtype == "float":
                value = float(value)
            elif dtype == "intmapping":
                value = int(value)
            elif dtype == "strmapping":
                value = str(value)
        except KeyError:
            pass
        cmd = serial.dumps(
            {"action": "set_control_value", "control_id": control_id, "value": value}
        )
        self.command_push.send_string(self.uuid, flags=zmq.SNDMORE)
        self.command_push.send_string(cmd)


SensorFetchDataValue = typing.TypeVar("SensorFetchDataValue")


class SensorFetchDataMixin(typing.Generic[SensorFetchDataValue], abc.ABC):
    @property
    @abc.abstractmethod
    def formatter(self) -> DataFormatter[SensorFetchDataValue]:
        pass

    def fetch_data(self) -> typing.Iterator[SensorFetchDataValue]:
        assert isinstance(self, Sensor)

        if not self.supports_data_subscription:
            raise NotDataSubSupportedError()

        while self.has_data:
            data_msg = self.get_data(copy=False)
            data_msg = DataMessage(*data_msg)
            yield from self.formatter.decode_msg(data_msg=data_msg)


class VideoSensor(SensorFetchDataMixin[VideoValue], Sensor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._recent_frame = None
        self._waiting_for_iframe = True
        self._formatter = VideoDataFormatter.get_formatter(format=self.format)

    @property
    def formatter(self) -> VideoDataFormatter:
        return self._formatter

    def get_newest_data_frame(self, timeout=None):
        if not self.supports_data_subscription:
            raise NotDataSubSupportedError()

        if self.data_sub.poll(timeout=timeout):
            newest_frame = None
            for newest_frame in self.fetch_data():
                # Get the last avaiable frame
                pass
            if newest_frame is not None:
                return newest_frame
            else:
                raise StreamError("Operation timed out.")
        else:
            raise StreamError("Operation timed out.")


class AnnotateSensor(SensorFetchDataMixin[AnnotateValue], Sensor):
    @property
    def formatter(self) -> AnnotateDataFormatter:
        return AnnotateDataFormatter.get_formatter(format=self.format)

    def _init_data_sub(self, context):
        if self.data_endpoint:
            self.data_sub = context.socket(zmq.SUB)
            self.data_sub.set_hwm(3)
            self.data_sub.connect(self.data_endpoint)
            self.data_sub.subscribe("")
        else:
            self.data_sub = None


class GazeSensor(SensorFetchDataMixin[GazeValue], Sensor):
    @property
    def formatter(self) -> GazeDataFormatter:
        return GazeDataFormatter.get_formatter(format=self.format)


class IMUSensor(SensorFetchDataMixin[IMUValue], Sensor):
    @property
    def formatter(self) -> IMUDataFormatter:
        return IMUDataFormatter.get_formatter(format=self.format)


class EventSensor(SensorFetchDataMixin[EventValue], Sensor):
    @property
    def formatter(self) -> EventDataFormatter:
        return EventDataFormatter.get_formatter(format=self.format)


class LEDSensor(Sensor):
    pass


_SENSOR_TYPE_CLASS_MAP = {
    SensorType.HARDWARE: Sensor,
    SensorType.VIDEO: VideoSensor,
    SensorType.ANNOTATE: AnnotateSensor,
    SensorType.GAZE: GazeSensor,
    SensorType.IMU: IMUSensor,
    SensorType.EVENT: EventSensor,
    SensorType.LED: LEDSensor,
}



"""
(*)~----------------------------------------------------------------------------------
 Pupil - eye tracking platform
 Copyright (C) 2012-2015  Pupil Labs

 Distributed under the terms of the CC BY-NC-SA License.
 License details are in the file license.txt, distributed as part of this software.
----------------------------------------------------------------------------------~(*)
"""
import glob
import os
import pathlib
import platform

import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup

requirements = [
    "numpy",
    "pyzmq",
    "zeromq-pyre",
]

here = pathlib.Path(__file__).parent
with open(here / "README.md") as f:
    long_description = f.read()


libs = []
library_dirs = []
include_dirs = []
extra_link_args = []
extra_objects = []
include_dirs = [numpy.get_include()]
if os.environ.get("CIBUILDWHEEL"):
    if platform.system() == "Windows":
        ffmpeg_base = "C:\\cibw\\vendor\\"
        libjpeg_turbo_base = "C:\\cibw\\libjpeg-turbo-build\\"
        include_dirs += [
            f"{libjpeg_turbo_base}include",
            f"{ffmpeg_base}include",
            "ndsi\\h264\\windows",
        ]
        extra_objects += [f"{libjpeg_turbo_base}lib\\turbojpeg-static.lib"]
        libs = ["winmm"] + [
            f"{ffmpeg_base}lib\\{L}"
            for L in ("avutil", "avformat", "avcodec", "swscale")
        ]
    else:
        include_dirs += ["/tmp/libjpeg-turbo-build/include/", "/tmp/vendor/include/"]
        library_dirs += ["/tmp/libjpeg-turbo-build/lib/", "/tmp/vendor/lib/"]
        for folder in include_dirs + library_dirs:
            assert os.path.exists(folder), f"{folder} not found!"
        libs += ["turbojpeg", "avutil", "avformat", "avcodec", "swscale"]
elif platform.system() == "Darwin":
    include_dirs += ["/usr/local/opt/jpeg-turbo/include/"]
    libs += ["turbojpeg"]
    library_dirs += ["/usr/local/opt/jpeg-turbo/lib/"]
    libs += ["avutil", "avformat", "avcodec", "swscale"]
elif platform.system() == "Linux":
    libs = ["rt", "turbojpeg"]
    libs += ["avutil", "avformat", "avcodec", "swscale"]
elif platform.system() == "Windows":
    libs = ["winmm"]
    tj_dir = "C:\\work\\libjpeg-turbo-VC64"
    tj_lib = tj_dir + "\\lib\\turbojpeg.lib"
    include_dirs += [tj_dir + "\\include"]
    extra_objects += [tj_lib]
    include_dirs += ["ndsi\\h264\\windows"]
    ffmpeg_libs = "C:\\work\\ffmpeg-4.0-win64-dev\\lib"
    include_dirs += ["C:\\work\\ffmpeg-4.0-win64-dev\\include"]
    libs += [
        ffmpeg_libs + "\\avutil",
        ffmpeg_libs + "\\avformat",
        ffmpeg_libs + "\\avcodec",
        ffmpeg_libs + "\\swscale",
    ]

h264_sources = glob.glob("ndsi/h264/*.cpp")
if platform.system() == "Windows":
    h264_sources += glob.glob("ndsi/h264/windows/*.cpp")

extensions = [
    Extension(
        name="ndsi.frame",
        sources=h264_sources + ["ndsi/frame.pyx"],
        include_dirs=[numpy.get_include()] + include_dirs,
        library_dirs=library_dirs,
        libraries=libs,
        extra_link_args=extra_link_args + ["-std=c++11"],
        extra_compile_args=["-std=c++11"],
        extra_objects=extra_objects,
        language="c++",
    ),
    Extension(
        name="ndsi.writer",
        sources=h264_sources + ["ndsi/writer.pyx"],
        include_dirs=[numpy.get_include()] + include_dirs,
        library_dirs=library_dirs,
        libraries=libs,
        extra_link_args=extra_link_args + ["-std=c++11"],
        extra_compile_args=["-std=c++11"],
        extra_objects=extra_objects,
        language="c++",
    ),
]

setup(
    name="ndsi",
    version="1.4.2",
    install_requires=requirements,
    extras_require={
        # TODO: Publish pyuvc via PyPI and reenable:
        # "examples": ["uvc"],
        "dev": ["pytest", "bump2version", "black"],
    },
    description="Remote Device Sensor Interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["ndsi"],
    ext_modules=cythonize(extensions),
    url="https://github.com/pupil-labs/pyndsi",
    author="Pupil Labs",
    author_email="info@pupil-labs.com",
    license="LGPL-3.0",
    classifiers={
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: System :: Networking",
    },
    project_urls={"Changelog": "https://github.com/pupil-labs/pyndsi#Changelog"},
)
